import os
import random
from pprint import pprint

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from sklearn.metrics import classification_report, confusion_matrix, roc_auc_score, roc_curve
from torch.utils import data
from torchinfo import summary

from models import resnet_attention_gate_aspp
from utils.utils import TensorDataset


def setup_seed(seed=None, workers=False):
    os.environ["PL_GLOBAL_SEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    os.environ["PL_SEED_WORKERS"] = f"{int(workers)}"

    return seed


def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2 ** 32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


def adjust_learning_rate(optimizer, epoch, learning_rate):
    lr_new = learning_rate * (0.1 ** (epoch // 10))
    print("lr:", lr_new)
    for param_group in optimizer.param_groups:
        param_group["lr"] = lr_new


def train(train_loader, model, criterion, optimizer, epoch, device):
    model.train()

    sum_loss = 0
    correct = 0
    total_num = len(train_loader.dataset)
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        output = model(data)
        classify_loss = criterion(output, target)

        regularization_loss = 0
        for param in model.parameters():
            # regularization_loss += torch.sum(torch.pow(param, 2.0))
            regularization_loss += torch.sum(torch.abs(param))

        loss = classify_loss + regularization_loss * 1e-2

        pred = output.data.argmax(dim=1, keepdim=True)
        correct += pred.eq(target.view_as(pred)).sum().item()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        print_loss = loss.data.item()
        sum_loss += print_loss
        if (batch_idx + 1) % int(len(train_loader) / 10) == 0:
            print("Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                epoch, (batch_idx + 1) * len(data), len(train_loader.dataset),
                       100. * (batch_idx + 1) / len(train_loader), loss.data.item()
            ))

    avg_loss = sum_loss / len(train_loader)
    acc = correct / total_num
    print("epoch: {}, loss: {:.4f}, accuracy: {:.1f}".format(epoch, avg_loss, 100 * acc))

    return acc, avg_loss


def val(validation_loader, model, criterion, device):
    model.eval()

    test_loss = 0
    correct = 0
    total_num = len(validation_loader.dataset)
    with torch.no_grad():
        for data, target in validation_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            classify_loss = criterion(output, target)

            regularization_loss = 0
            for param in model.parameters():
                # regularization_loss += torch.sum(torch.pow(param, 2.0))
                regularization_loss += torch.sum(torch.abs(param))

            loss = classify_loss + regularization_loss * 1e-2

            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).cpu().sum()

            print_loss = loss.data.item()
            test_loss += print_loss

    avg_loss = test_loss / len(validation_loader)
    acc = correct / total_num
    print("\nVal set: Average loss: {:.4f}, Accuracy: {}/{} ({:.1f}%)\n".format(
        avg_loss, correct, len(validation_loader.dataset), 100 * acc
    ))

    return acc, avg_loss


def test(test_loader, model, device):
    model.eval()

    y_score, y_true = [], []
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)

            y_score.append(F.softmax(output, dim=1).cpu().numpy())
            y_true.append(target.cpu().numpy())

    y_score = np.concatenate(y_score, axis=0)
    y_true = np.concatenate(y_true, axis=0)

    return y_score, y_true


def read_data(random_state=0):
    data = np.load("data/mci_data.npz")
    X_train, y_train, X_test, y_test = data["X_train"], data["y_train"], data["X_test"], data["y_test"]
    X_train = np.expand_dims(X_train, axis=1)
    X_test = np.expand_dims(X_test, axis=1)

    print("train data size:", X_train.shape)
    print("test data size:", X_test.shape)

    # numpy --> tensor
    X_train, y_train = torch.tensor(X_train, dtype=torch.float32), torch.tensor(y_train, dtype=torch.int64)
    X_test, y_test = torch.tensor(X_test, dtype=torch.float32), torch.tensor(y_test, dtype=torch.int64)

    return X_train, y_train, X_test, y_test


def main():
    random_state = 0
    setup_seed(random_state)

    g = torch.Generator()
    g.manual_seed(random_state)

    batch_size = 8
    epochs = 50
    learning_rate = 0.001
    device = "cuda" if torch.cuda.is_available() else "cpu"

    X_train, y_train, X_test, y_test = read_data(random_state=random_state)

    dataset_train = TensorDataset(X_train, y_train)
    dataset_test = TensorDataset(X_test, y_test)

    train_loader = data.DataLoader(dataset_train, batch_size=batch_size, shuffle=True, num_workers=5,
                                   worker_init_fn=seed_worker, generator=g)
    test_loader = data.DataLoader(dataset_test, batch_size=batch_size, shuffle=False)

    model = resnet_attention_gate_aspp.resnet18()
    summary(model, input_size=(batch_size, 1, 42, 51, 42))
    model.to(device)

    model_path = "saved_model/ag_resnet.pth"
    log_path = "log/ag_resnet.npz"
    if not os.path.exists(model_path):
        criterion = nn.CrossEntropyLoss()
        optimizer = optim.Adam(model.parameters(), lr=learning_rate)

        accs, losses = [], []
        val_accs, val_losses = [], []
        for epoch in range(1, epochs + 1):
            adjust_learning_rate(optimizer, epoch, learning_rate)
            acc, loss = train(train_loader, model, criterion, optimizer, epoch, device)
            val_acc, val_loss = val(test_loader, model, criterion, device)

            accs.append(acc)
            losses.append(loss)
            val_accs.append(val_acc)
            val_losses.append(val_loss)

        torch.save(model, model_path)
        np.savez(log_path, losses=losses, val_losses=val_losses, accs=accs, val_accs=val_accs)
    else:
        model = torch.load(model_path)
        log_data = np.load(log_path)

        losses, val_losses, accs, val_accs = log_data["losses"], log_data["val_losses"], \
            log_data["accs"], log_data["val_accs"]

    y_score, y_true = test(test_loader, model, device)
    y_pred = np.argmax(y_score, axis=1)

    tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
    sn, sp, acc, f1 = tp / (tp + fn), tn / (fp + tn), (tn + tp) / (tn + fp + fn + tp), 2 * tp / (2 * tp + fp + fn)

    fpr, tpr, _ = roc_curve(y_true, y_score[:, 1], pos_label=1)
    auc = roc_auc_score(y_true, y_score[:, 1])

    print("#Sn\tSp\tAcc\tAUC\tF1")
    print("{:.3f}\t{:.3f}\t{:.3f}\t{:.3f}\t{:.3f}\n".format(sn, sp, acc, auc, f1))

    print("Test set:")
    print(classification_report(y_true, y_pred, target_names=["Normal", "MCI"], digits=3))
    pprint(confusion_matrix(y_true, y_pred))

    fig, axes = plt.subplots(1, 2, figsize=(10, 4.5))
    axes[0].plot(losses, label="loss")
    axes[0].plot(val_losses, label="val_loss")
    axes[0].set(xlabel="Epoch", title="Loss")
    axes[0].legend(loc="upper right")

    axes[1].plot(accs, label="acc")
    axes[1].plot(val_accs, label="val_acc")
    axes[1].set(xlabel="Epoch", title="Accuracy")
    axes[1].legend(loc="lower right")
    fig.tight_layout()
    plt.show()


if __name__ == '__main__':
    main()
